# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END

""" Plugin module to forecast the possibility of breakover. """


__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


import filter_ingest
from foglamp.common.bucket_client.bucket_client import BucketClient
from foglamp.plugins.filter.breakover.core.base import _LOGGER
from foglamp.plugins.filter.breakover.config import PLUGIN_NAME, DEFAULT_CONFIG
from foglamp.plugins.filter.breakover.exec import BreakoverExec


def plugin_info():
    """Return information about the plugin.

    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.debug(
        "Inside {} plugin_info(); Returning the plugin info details".format(
            PLUGIN_NAME
        )
    )
    return {
        "name": PLUGIN_NAME,
        "version": "3.0.0",
        "mode": "none",
        "type": "filter",
        "interface": "1.0",
        "config": DEFAULT_CONFIG,
    }


def plugin_init(config, ingest_ref, callback):
    """Initialise the plugin.

    Args:
        config      : JSON configuration document for the Filter plugin
                      configuration category
        ingest_ref  : filter ingest reference
        callback    : filter callback
    Returns:
        handle      : JSON object to be used in future calls to the plugin
    Raises:
    """
    global _LOGGER

    _LOGGER.debug("Inside {} plugin_init()".format(PLUGIN_NAME))

    handle = {}
    try:
        if (
            ("bucket_client" not in config) or
            (not isinstance(config["bucket_client"], BucketClient))
        ):
            url = config["mgmt_client_url_base"]["value"]
            host, port = url.split(':')
            _LOGGER.info("urlBase={}, host={}, port={}".format(url, host, port))
            handle["mgmt_client_url_base"] = url
            handle["bucket_client"] = BucketClient(host, port)
        else:
            _LOGGER.info("BucketClient is already instantiated")
            handle["mgmt_client_url_base"] = config["mgmt_client_url_base"]
            handle["bucket_client"] = config["bucket_client"]
    except Exception as ex:
        _msg  = "Breakover-Filter is unable to connect to Bucket service, "
        _msg += f"restart plugin after it becomes available. Details: {str(ex)}"
        _LOGGER.error(_msg)
        handle["bucket_client"] = None
        if config["model_path"]["value"] is None:
            raise Exception(_msg)

    handle["breakover"] = BreakoverExec(config, handle["bucket_client"])

    _LOGGER = handle["breakover"].get_logger()
    _LOGGER.debug("\n\tconfig: " + str(config))

    handle["ingestRef"] = ingest_ref
    handle["callback"] = callback

    _LOGGER.debug("\n\thandle created (updated config): " + str(handle))
    return handle


def plugin_reconfigure(handle, new_config):
    """Reconfigure the plugin.

    Args:
        handle     : handle returned by the plugin initialization call
        new_config : JSON object representing the new configuration category for
                     the category
    Returns:
        new_handle : new handle to be used in the future calls
    """
    _LOGGER.debug("Inside {} plugin_reconfigure()".format(PLUGIN_NAME))
    _LOGGER.debug(
        "Reconfiguring {}; \n\tOld Config: {} \n\tNew Config: {}".format(
            PLUGIN_NAME, handle, new_config
        )
    )

    if (
        ("mgmt_client_url_base" in handle) and ("bucket_client" in handle)
    ):
        new_config["mgmt_client_url_base"] = handle["mgmt_client_url_base"]
        new_config["bucket_client"] = handle["bucket_client"]

    new_handle = plugin_init(new_config, handle["ingestRef"], handle["callback"])

    return new_handle


def plugin_shutdown(handle):
    """Shutdown the plugin doing required cleanup.

    Args:
        handle: handle returned by the plugin initialization call
    Returns:
    """
    _LOGGER.debug("Inside {} plugin_shutdown()".format(PLUGIN_NAME))

    handle["callback"] = None
    handle["ingestRef"] = None

    handle["breakover"].stop()
    handle["breakover"] = None

    _LOGGER.info("{} filter plugin shutdown completed.".format(PLUGIN_NAME))


def plugin_ingest(handle, data):
    """Modify readings data and pass it onward.

    Args:
        handle: handle returned by the plugin initialization call
        data: readings data
    """
    _LOGGER.debug("Inside {} plugin_ingest()".format(PLUGIN_NAME))
    _LOGGER.debug("handle:   " + str(handle))
    _LOGGER.debug("Incoming Data  :   " + str(type(data)) + " " + str(data))

    process_data = None
    try:
        process_data = handle["breakover"].run(data)
    except Exception as e:
        _LOGGER.error("Error in `breakover` inference plugin: " + str(e))

    # Caution: allow forwarding the data even in case of failure.
    # Pass data onwards: if Filter is not enabled then process_data is None
    # (designed this way) in that case send unprocessed data onwards;
    outgoing_data = data if process_data is None else process_data
    filter_ingest.filter_ingest_callback(
        handle["callback"],
        handle["ingestRef"],
        outgoing_data,
    )

    _LOGGER.debug("Outgoing Data  :   " + str(outgoing_data))
    _LOGGER.debug("{} filter ingestion performed".format(PLUGIN_NAME))

    if isinstance(data, dict):
        data.clear()
    if isinstance(process_data, dict):
        process_data.clear()
    if isinstance(outgoing_data, dict):
        outgoing_data.clear()

    del data, process_data, outgoing_data
