# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END


__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


import json
import pandas as pd

from foglamp.plugins.filter.breakover.core.base import _LOGGER
from foglamp.plugins.filter.breakover.core.bucket_handler import (
    bucket_config_properties, bucket_config_default
)
from foglamp.plugins.filter.breakover.core.post_process_handler import PostProcess


PLUGIN_NAME = "breakover"
_grp1 = "Input"
_grp2 = "Inference"
_grp3 = "Post Process"
_grp4 = "Output"
_grp5 = "Enable"


_DEFAULT_INPUT = [
    {
        "asset": "asset_1",
        "datapoint": "datapoint_1",
        "feature_type": "Boolean",
        "boolean_data_scaling": "None",
        "numeric_data_scaling": "None",
        "min_value": "-1.0",
        "median_value": "-1.0",
        "mean_value": "-1.0",
        "max_value": "-1.0",
        "iqr_value": "-1.0",
        "std_value": "-1.0"
    }
]

_DEFAULT_POST_PROCESS = [
    {
        "state_name": "healthy",
        "state_id": "0",
        "technique": "Score above threshold",
        "threshold": "0.0",
        "rolling_window_length": "1",
        "rolling_measure_name": "Average",
        "duration_threshold": "1"
    },
    {
        "state_name": "warning",
        "state_id": "1",
        "technique": "Rolling measure above threshold",
        "rolling_window_length": "7",
        "rolling_measure_name": "Median",
        "threshold": "0.7",
        "duration_threshold": "1"
    },
    {
        "state_name": "critical",
        "state_id": "2",
        "technique": "Rolling measure above threshold",
        "rolling_window_length": "10",
        "rolling_measure_name": "Median",
        "threshold": "0.9",
        "duration_threshold": "1"
    }
]

DEFAULT_CONFIG = {
    "plugin": {
        "group": "Enable",
        "description": "Filter for breakover forecasting.",
        "type": "string",
        "default": PLUGIN_NAME,
        "readonly": "true",
    },
    "bucket_config": {
        "order": "1",
        "displayName": "ML Model",
        "description": "Breakover forecasting model to be imported from bucket.",
        "type": "bucket",
        "properties": bucket_config_properties,
        "default": json.dumps(bucket_config_default)
    },
    "inputs": {
        "order": "2",
        "group": _grp1,
        "displayName": "Inputs",
        "description": "Configure input data that is to be used for breakover forecasting.",
        "type": "list",
        "items": "object",
        "default": json.dumps(_DEFAULT_INPUT),
        "properties": {
            "asset": {
                "displayName": "Asset name",
                "description": "Enter input asset name.",
                "type": "string",
                "default": "asset_1",
            },
            "datapoint": {
                "displayName": "Datapoint name",
                "description": "Enter input datapoint name.",
                "type": "string",
                "default": "datapoint_1",
            },
            "feature_type": {
                "displayName": "Feature type",
                "description": "Datatype of the feature.",
                "type": "enumeration",
                "options": [
                    "Boolean",
                    "Numeric",
                ],
                "default": "Boolean",
            },
            "boolean_data_scaling": {
                "validity": 'feature_type == "Boolean"',
                "displayName": "Boolean scaling strategy",
                "description": "The strategy that is to be used for data scaling of boolean feature.",
                "type": "enumeration",
                "options": [
                    "None",
                    "Custom Encoded",
                ],
                "default": "None",
            },
            "numeric_data_scaling": {
                "validity": 'feature_type == "Numeric"',
                "displayName": "Numeric scaling strategy",
                "description": "The strategy that is to be used for data scaling of numeric feature.",
                "type": "enumeration",
                "options": [
                    "None",
                    "Custom Robust",
                    "Normalization",
                    "Robust",
                    "Standardization",
                ],
                "default": "None",
            },
            "min_value": {
                "validity": 'feature_type == "Numeric" && ( numeric_data_scaling == "Custom Robust" || numeric_data_scaling == "Normalization" )',
                "displayName": "Minimum",
                "description": "Enter minimum value for data rescaling.",
                "type": "float",
                "default": "-1.0",
            },
            "median_value": {
                "validity": 'feature_type == "Numeric" &&  ( numeric_data_scaling == "Custom Robust" || numeric_data_scaling == "Robust" )',
                "displayName": "Median",
                "description": "Enter median value for data rescaling.",
                "type": "float",
                "default": "-1.0",
            },
            "mean_value": {
                "validity": 'feature_type == "Numeric" && numeric_data_scaling == "Standardization"',
                "displayName": "Mean",
                "description": "Enter mean value for data rescaling.",
                "type": "float",
                "default": "-1.0",
            },
            "max_value": {
                "validity": 'feature_type == "Numeric" && ( numeric_data_scaling == "Custom Robust" || numeric_data_scaling == "Normalization" )',
                "displayName": "Maximum",
                "description": "Enter maximum value for data rescaling.",
                "type": "float",
                "default": "-1.0",
            },
            "iqr_value": {
                "validity": 'feature_type == "Numeric" && numeric_data_scaling == "Robust"',
                "displayName": "Interquartile range",
                "description": "Enter interquartile range value for data rescaling.",
                "type": "float",
                "default": "-1.0",
            },
            "std_value": {
                "validity": 'feature_type == "Numeric" && numeric_data_scaling == "Standardization"',
                "displayName": "Standard deviation",
                "description": "Enter standard deviation value for data rescaling.",
                "type": "float",
                "default": "-1.0",
            }
        }
    },
    "sequence_length": {
        "order": "3",
        "group": _grp2,
        "displayName": "Sequence size",
        "description": "Sequence size that is to be used for breakover forecasting.",
        "type": "integer",
        "minimum": "10",
        "maximum": "300",
        "default": "10",
    },
    "inference_interval": {
        "order": "4",
        "group": _grp2,
        "displayName": "Inference interval",
        "description": "Time interval after which the data will be passed "
                        "through the model.",
        "type": "float",
        "default": "1.0",
    },
    "inference_interval_unit": {
        "order": "5",
        "group": _grp2,
        "displayName": "Inference interval unit",
        "description": "Time interval unit",
        "type": "enumeration",
        "options": [
            "seconds",
            "minutes",
            "hours",
        ],
        "default": "seconds",
    },
    "post_process": {
        "order": "6",
        "group": _grp3,
        "displayName": _grp3,
        "description": "The post processing config for the score.",
        "type": "list",
        "items": "object",
        "default": json.dumps(_DEFAULT_POST_PROCESS),
        "properties": {
            "state_name": {
                "displayName": "State name",
                "description": "Enter the state name.",
                "type": "string",
                "default": "healthy",
            },
            "state_id": {
                "displayName": "State ID",
                "description": "Enter the state ID.",
                "type": "integer",
                "default": "0",
            },
            "technique": {
                "displayName": "Technique",
                "description": "Select the post processing technique to be used.",
                "type": "enumeration",
                "options": list(PostProcess.techniques.keys()),
                "default": "Score above threshold",
            },
            "rolling_window_length": {
                "validity": 'technique == "Rolling measure above threshold" || technique == "Rolling measure above threshold for x duration"',
                "displayName": "Rolling window length",
                "description": "Number of last score values that is to be used for breakover forecasting.",
                "type": "integer",
                "minimum": "1",
                "maximum": "300",
                "default": "1",
            },
            "rolling_measure_name": {
                "validity": 'technique == "Rolling measure above threshold" || technique == "Rolling measure above threshold for x duration"',
                "displayName": "Rolling measure name",
                "description": "Measure to calculate on the scores value present in the window.",
                "type": "enumeration",
                "options": list(PostProcess.measures.keys()),
                "default": "Average",
            },
            "threshold": {
                "displayName": "Threshold",
                "description": "Threshold to apply on the calculated measure value.",
                "type": "float",
                "default": "0.5",
            },
            "duration_threshold": {
                "validity": 'technique == "Score above threshold for x duration" || technique == "Rolling measure above threshold for x duration"',
                "displayName": "Duration cutoff",
                "description": "The amount of time the post processed score should remain above threshold (in seconds).",
                "type": "integer",
                "default": "1",
            },
        }
    },
    "forward_data": {
        "order": "7",
        "group": _grp4,
        "displayName": "Forward data",
        "description": "Whether to forward received PLC data down the filter chain.",
        "type": "boolean",
        "default": "true",
    },
    "output_asset_name": {
        "order": "8",
        "group": _grp4,
        "displayName": "Asset name",
        "description": "Name of asset that indicates breakover forecast.",
        "type": "string",
        "default": "breakover",
    },
    "output_score_datapoint": {
        "order": "9",
        "group": _grp4,
        "displayName": "Score datapoint",
        "description": "Name of datapoint that indicates breakover forecast score.",
        "type": "string",
        "default": "forecast_score",
    },
    "output_state_wise_forecast": {
        "order": "10",
        "group": _grp4,
        "displayName": "Add state wise datapoint",
        "description": "Add separate datapoint for state wise forecast.",
        "type": "boolean",
        "default": "false",
    },
    "output_state_wise_forecast_datapoint_name_ptrn": {
        "order": "11",
        "group": _grp4,
        "validity": 'output_state_wise_forecast == "true"',
        "displayName": "Prefix for state wise datapoint",
        "description": "Prefix of datapoint name that indicates state wise breakover forecast.",
        "type": "string",
        "default": "forecast_imminent_",
    },
    "output_forecast_state_datapoint": {
        "order": "12",
        "group": _grp4,
        "displayName": "Forecasted state datapoint",
        "description": "Name of datapoint that indicates state/level of the breakover forecast.",
        "type": "string",
        "default": "forecasted_state",
    },
    "output_return_state": {
        "order": "13",
        "group": _grp4,
        "displayName": "Return forecasted state",
        "description": "Whether to return forecasted state name or state ID.",
        "type": "enumeration",
        "options": ["State name", "State ID"],
        "default": "State ID",
    },
    "add_metadata": {
        "order": "14",
        "group": _grp5,
        "displayName": "Add performance counters",
        "description": "Add performance counters metadata containing resource utilization statistics.",
        "type": "boolean",
        "default": "false",
    },
    "log_level": {
        "order": "15",
        "group": _grp5,
        "displayName": "Log level",
        "description": "Log level for diagnostic or error output",
        "type": "enumeration",
        "options": ["ERROR", "WARN", "INFO", "DEBUG"],
        "default": "INFO",
    },
    "enable": {
        "order": "16",
        "group": _grp5,
        "displayName": "Enabled",
        "description": "Enable breakover forecast filter plugin",
        "type": "boolean",
        "default": "false",
    },
}


def _create_input_dict(info):
    """Cleanup the input config object.

    Args:
        info: single config object part of input list
    Returns:
        restructured_config: dict eg. below
            {
                "asset": "asset_1",            # mandatory
                "datapoint": "datapoint_1",    # mandatory
                "feature_type": "boolean",     # mandatory; options: `boolean`, `numeric`
                "scaler_strategy": None        # mandatory; options: None, `custom robust`, `normalization`, `robust`, `standardization`
                "minimum": 0.0,                # optional; with `numeric & scaler_strategy in [custom robust, normalization]`
                "maximum": 100.0,              # optional; with `numeric & scaler_strategy in [custom robust, normalization]`
                "mean": 50.0,                  # optional; with `numeric & scaler_strategy==standardization`
                "median": 50.0,                # optional; with `numeric & scaler_strategy in [custom robust, robust]`
                "iqr": 50.0,                   # optional; with `numeric & scaler_strategy==robust`
                "std": 0.5,                    # optional; with `numeric & scaler_strategy==standardization`
            }
    """
    feature_type = info["feature_type"].lower()
    boolean_scaler_strategy = (
        info["boolean_data_scaling"].lower() if "boolean_data_scaling" in info else "none"
    )
    numeric_scaler_strategy = (
        info["numeric_data_scaling"].lower() if "numeric_data_scaling" in info else "none"
    )
    if feature_type == "boolean":
        scaler_strategy = boolean_scaler_strategy
    elif feature_type == "numeric":
        scaler_strategy = numeric_scaler_strategy
    else:
        _msg  = f"The provided feature type `{feature_type}` isn't supported."
        _LOGGER.error(_msg)

    del boolean_scaler_strategy, numeric_scaler_strategy

    supported_combination = {
        "boolean": ["none", "custom encoded"],
        "numeric": [
            "none", "custom robust", "normalization", "robust", "standardization"
        ],
    }
    if scaler_strategy not in supported_combination[feature_type]:
        _msg  = f"The provided scaling strategy `{scaler_strategy}` isn't "
        _msg += f"supported for `{feature_type}` feature type."
        _LOGGER.error(_msg)
        raise Exception(_msg)

    # Creating config
    d = {
        "asset": info["asset"],
        "datapoint": info["datapoint"],
        "feature_type": feature_type,
        "scaler_strategy": scaler_strategy if scaler_strategy != "none" else None
    }
    if d["scaler_strategy"] == "custom robust":
        d["stats"] = {
            "median": float(info["median_value"]),
            "minimum": float(info["min_value"]),
            "maximum": float(info["max_value"])
        }
    elif d["scaler_strategy"] == "normalization":
        d["stats"] = {
            "minimum": float(info["min_value"]),
            "maximum": float(info["max_value"])
        }
    elif d["scaler_strategy"] == "robust":
        d["stats"] = {
            "median": float(info["median_value"]),
            "iqr": float(info["iqr_value"])
        }
    elif d["scaler_strategy"] == "standardization":
        d["stats"] = {
            "mean": float(info["mean_value"]),
            "std": float(info["std_value"])
        }
    return d


def _create_post_process_dict(info):
    """Cleanup the post process config object.

    Args:
        info: single config object part of post process list
    Returns:
        restructured config for post process element: dict
    """
    rolling_techniques = [
        "Rolling measure above threshold",
        "Rolling measure above threshold for x duration"
    ]
    duration_cutoff_technique = [
        "Score above threshold for x duration",
        "Rolling measure above threshold for x duration"
    ]
    technique = info["technique"]

    d = {
        "state_name": info["state_name"],
        "state_id": int(info["state_id"]),
        "technique": technique,
        "rolling_window_length": (
            int(info["rolling_window_length"])
            if ((technique in rolling_techniques) and ("rolling_window_length" in info)) else
            None
        ),
        "rolling_measure_name": (
            info["rolling_measure_name"]
            if ((technique in rolling_techniques) and ("rolling_measure_name" in info)) else
            None
        ),
        "threshold": float(info["threshold"]),
        "duration_cutoff": (
            int(info["duration_cutoff"])
            if ((technique in duration_cutoff_technique) and ("duration_cutoff" in info)) else
            None
        )
    }
    return d


def parse_config(plugin_config):
    """Plugin configuration parser."""
    for key in ["inputs", "bucket_config", "post_process"]:
        plugin_config[key]["value"] = json.loads(plugin_config[key]["value"])

    breakover_config = {
        "inputs": [
            _create_input_dict(info)
            for info in plugin_config["inputs"]["value"]
        ],
        "inference": {
            "sequence_length": int(plugin_config["sequence_length"]["value"]),
            "trigger_interval": pd.to_timedelta("{} {}".format(
                plugin_config["inference_interval"]["value"],
                plugin_config["inference_interval_unit"]["value"]
            )).total_seconds(),
            "bucket": {
                "logger": None,
                "bucket_client": None,
                "bucket_type": plugin_config["bucket_config"]["value"]["type"],
                "model_name": plugin_config["bucket_config"]["value"]["name"],
                "model_version": plugin_config["bucket_config"]["value"]["version"],
            }
        },
        "post_process": [
            _create_post_process_dict(info)
            for info in plugin_config["post_process"]["value"]
        ],
        "output": {
            "asset": plugin_config["output_asset_name"]["value"],
            "score_datapoint": plugin_config["output_score_datapoint"]["value"],
            "state_wise_forecast_datapoint_prefix": (
                plugin_config["output_state_wise_forecast_datapoint_name_ptrn"]["value"]
                if plugin_config["output_state_wise_forecast"]["value"] == "true" else
                None
            ),
            "return_state": (
                "name"
                if plugin_config["output_return_state"]["value"] == "State name" else (
                    "id"
                    if plugin_config["output_return_state"]["value"] == "State ID" else
                    None
                )
            ),
            "forecasted_state_datapoint": (
                plugin_config["output_forecast_state_datapoint"]["value"]
                if plugin_config["output_return_state"]["value"] in ["State name", "State ID"] else
                None
            ),
            "forward_data": plugin_config["forward_data"]["value"] == "true",
        },
        "add_metadata": plugin_config["add_metadata"]["value"] == "true",
        "log_level": plugin_config["log_level"]["value"],
        "enable": plugin_config["enable"]["value"] == "true"
    }
    _LOGGER.debug(f"CUSTOM MESSAGE: `{str(type(breakover_config))}`  `{breakover_config}`")
    del plugin_config

    return breakover_config
