# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Plugin module to accumulate raw data and transform it on request basis. """

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


import time
import numpy as np
import pandas as pd

from foglamp.plugins.filter.breakover.core.base import _LOGGER


class DataAccumulator:
    """A class that accumulates raw data and transform it for forecasting on request basis."""

    __slots__ = [
        "allowed_asset_and_datapoints", "fields", "logger", "max_user_ts", "order_cols", 
        "overall_sequence_length", "raw_df", "sequence_length"
    ]

    def __init__(
        self,
        fields,
        sequence_length=120,
        pad_sequence_length=5,
        logger=_LOGGER
    ):
        """Initialize data accumulator.

        Args:
            fields: a list of dict with the following keys
                    - `asset`            : mandatory
                    - `datapoint`        : mandatory
                    - `feature_type`     : mandatory
                                         : options: `boolean`, `numeric`
                    - `scaler_strategy`  : mandatory
                                           options: None, `custom robust`, `normalization`, `robust`, `standardization`
                    - `minimum`          : optional; with `numeric & scaler_strategy in [custom robust, normalization]`
                    - `maximum`          : optional; with `numeric & scaler_strategy in [custom robust, normalization]`
                    - `mean`             : optional; with `numeric & scaler_strategy==standardization`
                    - `median`           : optional; with `numeric & scaler_strategy in [custom robust, robust]`
                    - `iqr`              : optional; with `numeric & scaler_strategy==robust`
                    - `std`              : optional; with `numeric & scaler_strategy==standardization`

            sequence_length: length of the sequence to create for forecasting
            pad_sequence_length: An extra sequence to be added for optimal data interpolation.
            logger: logger instance to use

        """
        self.fields = fields
        self.allowed_asset_and_datapoints = [
            {
                "asset": d["asset"],
                "datapoint": d["datapoint"],
            }
            for d in fields
        ]
        self.order_cols = [
            "{}_{}".format(d["asset"], d["datapoint"])
            for d in self.allowed_asset_and_datapoints
        ]

        self.raw_df = pd.DataFrame(
            columns=["user_ts", "asset_datapoint", "value"]
        )
        self.sequence_length = sequence_length
        self.overall_sequence_length = sequence_length + pad_sequence_length
        self.logger = logger
        self.max_user_ts = None

    def filter_and_add_data(self, data=[]):
        """Filter the data on asset and datapoint, then accumulate the filtered data.

        Args:
            data: readings data

            eg.
                data = [
                    {'readings': {'Current': 75.56722364251753, 'Backpressure': 43.054769951373274, 'Turbidity': 2.5430244703962055, 'LargeDischarge': 1, 'Status': 0, 'InletValveOpen': 1, 'ValveWaterFill': 1, 'ValveWaterFlush': 1, 'ValveAirFillMetering': 1, 'ValveHoodSpray': 0, 'ValveBowlOpen': 0, 'ValveBowlClose': 0}, 'asset': 'plc_A', 'id': 0, 'ts': '2024-09-11 11:42:46.132946+00:00', 'user_ts': '2024-09-11 11:42:46.132946+00:00'},
                    {'readings': {'SlipSpeed': 0, 'SmallDischarge': 0, 'InFeedOilThroughput': 321.8492001401262, 'InFeedWaterThroughput': 116.8129970344782, 'InFeedCausticThroughput': 45.05047557645883, 'VibrationWarning': 0, 'VibrationAlarm': 1}, 'asset': 'plc_B', 'id': 0, 'ts': '2024-09-11 11:42:46.132946+00:00', 'user_ts': '2024-09-11 11:42:46.132946+00:00'},
                    {'readings': {'Current': 78.94050733140269, 'Backpressure': 43.77876835523623, 'Turbidity': 2.5430244703962055, 'LargeDischarge': 1, 'Status': 0, 'InletValveOpen': 1, 'ValveWaterFill': 1, 'ValveWaterFlush': 1, 'ValveAirFillMetering': 1, 'ValveHoodSpray': 0, 'ValveBowlOpen': 1, 'ValveBowlClose': 0}, 'asset': 'plc_A', 'id': 0, 'ts': '2024-09-11 11:42:47.132929+00:00', 'user_ts': '2024-09-11 11:42:47.132929+00:00'},
                    {'readings': {'SlipSpeed': 0, 'SmallDischarge': 1, 'InFeedOilThroughput': 321.8492001401262, 'InFeedWaterThroughput': 116.8129970344782, 'InFeedCausticThroughput': 47.740212444570474, 'VibrationWarning': 1, 'VibrationAlarm': 1}, 'asset': 'plc_B', 'id': 0, 'ts': '2024-09-11 11:42:47.132929+00:00', 'user_ts': '2024-09-11 11:42:47.132929+00:00'},
                ]
        Return:

        """
        if isinstance(data, dict):
            data = [data]

        if len(data) == 0:
            return

        data = [
            {
                "user_ts": pd.to_datetime(d["user_ts"]),
                "asset_datapoint": "{}_{}".format(d["asset"], datapoint),
                "value": float(value),
            }
            for d in data
            for datapoint, value in d["readings"].items()
            if {"asset": d["asset"], "datapoint": datapoint} in self.allowed_asset_and_datapoints
        ]

        if len(data) == 0:
            return

        # Get the recent data and discard old expired data
        self.max_user_ts = max(data, key=lambda x: x["user_ts"])["user_ts"]

        allowed_min_user_ts = self.max_user_ts - pd.to_timedelta(
            f"{self.overall_sequence_length} seconds"
        )
        self.raw_df = (
            self.raw_df
            .loc[self.raw_df["user_ts"] >= allowed_min_user_ts, :]
        )

        # Add the new data
        self.raw_df = pd.concat([self.raw_df, pd.DataFrame(data)])

        self.logger.debug(f"`{len(data)}` readings added to breakover input queue.")

        del data

    def _prepare_and_filter_data(self):
        """Prepare data at one second granularity and keep readings in time range."""
        if len(self.raw_df) == 0:
            return None

        transformed_df = (
            pd.concat([
                (
                    self.raw_df
                    .loc[self.raw_df["asset_datapoint"] == tag, :]
                    .drop(columns=["asset_datapoint"])
                    .rename(columns={"value": tag})
                )
                for tag in self.order_cols
            ], axis=0)
            .set_index("user_ts")
            .sort_index()
            .interpolate(method="time")
            .bfill()
            .reset_index(drop=False)
            .groupby([pd.Grouper(key="user_ts", freq="1 s")])
            .mean(numeric_only=True)
            .interpolate(method="time")
        )

        # filtering to keep only in time range (remove non-pad region)
        allowed_min_user_ts = transformed_df.index.max() - pd.to_timedelta(
            f"{self.sequence_length} seconds"
        )
        transformed_df = (
            transformed_df
            .loc[transformed_df.index >= allowed_min_user_ts, :]
        )

        return transformed_df

    def _rescale_data(self, transformed_df):
        """Rescale the data."""
        for i, col in enumerate(self.order_cols):

            self.logger.debug(
                "Data rescale: col=`{}`; feature_type=`{}`; scaler_strategy=`{}`".format(
                    col, self.fields[i]["feature_type"], self.fields[i]["scaler_strategy"] 
                )
            )

            if self.fields[i]["feature_type"] == "boolean":
                if self.fields[i]["scaler_strategy"] is None:
                    pass

                elif self.fields[i]["scaler_strategy"] == "custom encoded":
                    transformed_df[col] = transformed_df[col].map(
                        {0: -0.5, 1: 0.5}
                    )

            elif self.fields[i]["feature_type"] == "numeric":

                if self.fields[i]["scaler_strategy"] is None:
                    pass

                elif self.fields[i]["scaler_strategy"] == "custom robust":
                    median_value = self.fields[i]["stats"]["median"]
                    min_value = self.fields[i]["stats"]["minimum"]
                    max_value = self.fields[i]["stats"]["maximum"]

                    transformed_df[col] = (
                        (transformed_df[col] - median_value) /
                        (max_value - min_value)
                    )

                    del median_value, min_value, max_value

                elif self.fields[i]["scaler_strategy"] == "normalization":
                    min_value = self.fields[i]["stats"]["minimum"]
                    max_value = self.fields[i]["stats"]["maximum"]

                    transformed_df[col] = (
                        (transformed_df[col] - min_value) /
                        (max_value - min_value)
                    )

                    del min_value, max_value

                elif self.fields[i]["scaler_strategy"] == "robust":
                    median_value = self.fields[i]["stats"]["median"]
                    iqr_value = self.fields[i]["stats"]["iqr"]

                    transformed_df[col] = (
                        (transformed_df[col] - median_value) /
                        iqr_value
                    )

                    del median_value, iqr_value

                elif self.fields[i]["scaler_strategy"] == "standardization":
                    mean_value = self.fields[i]["stats"]["mean"]
                    std_value = self.fields[i]["stats"]["std"]

                    transformed_df[col] = (
                        (transformed_df[col] - mean_value) /
                        std_value
                    )

                    del mean_value, std_value

        return transformed_df

    def get_prepared_data(self):
        """Transform the data and return it for forecasting.

        Args:

        Returns:
            recent_row       : most recent values
            dataset_array    : a 3D numpy array which will be input for score creation
            max_user_ts      : the max user timestamp in the accumulated data
            preparation_time : data transformation time in seconds
        """
        t0 = time.time()

        # convert to 1 second granularity & retain data in time range of interest
        transformed_df = self._prepare_and_filter_data()

        if len(transformed_df) > 0:
            recent_row = transformed_df.loc[transformed_df.index[-1], :].to_dict()
        else:
            recent_row = {}

        if transformed_df is None:
            self.logger.info(
                "No records present in the raw filtered dataset. Please check "
                "the plugin configuration to see if the right pair of asset and"
                " datapoints are configured and their data is incoming. "
                "Skipping this iteration..."
            )
            return recent_row, None, self.max_user_ts, time.time() - t0

        # Rescale the data
        transformed_df = self._rescale_data(transformed_df)

        # check the sequence length
        if transformed_df.shape[0] <= self.sequence_length:
            self.logger.info(
                "Insufficient sequence length available for breakover forecasting. "
                f"Required: {self.sequence_length}; Available: {transformed_df.shape[0]}"
                ". Skipping..."
            )
            return recent_row, None, self.max_user_ts, time.time() - t0

        # converting to array
        total_rows = len(transformed_df)
        dataset_array = np.array([
            (
                transformed_df
                .iloc[total_rows - self.sequence_length:total_rows, :]
                .to_numpy()
            )
        ])
        self.logger.debug(f"Array Shape : {dataset_array.shape}")

        del transformed_df, total_rows

        self.logger.debug(f"Time Taken in  DataAccumulator.get_prepared_data(): {time.time()-t0}")
        return recent_row, dataset_array, self.max_user_ts, time.time() - t0
