# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Plugin module to load the ML model and perform predictions."""

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


import os
import time
import numpy as np
import pandas as pd

# Select the backend model framework to use. Options: `keras`, `tflite`
MODEL_FRAMEWORK = "tflite"
if MODEL_FRAMEWORK == "keras":
    from tensorflow.keras.models import load_model as model_loader
    # Approximately 270MB are consumed by just importing tf lib
elif MODEL_FRAMEWORK == "tflite":
    from tflite_runtime.interpreter import Interpreter as model_loader
    # from tensorflow.lite import Interpreter as model_loader

from foglamp.plugins.filter.breakover.core.base import _LOGGER
from foglamp.plugins.filter.breakover.core.base import DEFAULT_SCORE


class Inference:
    """A class that loads model and perform predictions."""

    __slots__ = ["logger", "model", "model_framework", "input_details", "output_details"]

    def __init__(
        self,
        model_path,
        logger = _LOGGER
    ):
        """Initialize Inference class.

        Args:
            model_path: path to the ML model.
            model_framework: model framework
                options: `keras`, `tflite`
            logger: logger instance to use

        """
        self.logger = _LOGGER
        self.model_framework = MODEL_FRAMEWORK

        try:
            (
                self.model, self.input_details, self.output_details
            ) = self.load_model(model_path)
        except Exception as ex:
            _msg = "Issue in loading breakover model from `{}`. Details: {}".format(
                os.path.split(model_path)[-1], str(ex)
            )
            self.logger.error(_msg)
            raise Exception(_msg)

    def load_model(self, model_path):
        """Load breakover forecasting model and return it.

        Args:
            model_path: path to the model file
        Return:
            model: loaded model
        """
        model, input_details, output_details = None, None, None

        if self.model_framework == "keras":
            model = model_loader(model_path, compile=False)

        elif self.model_framework == "tflite":
            # Load the TFLite model and allocate tensors.
            model = model_loader(model_path=model_path)
            model.allocate_tensors()

            # Get input and output tensors.
            input_details = model.get_input_details()[0]["index"]
            output_details = model.get_output_details()[0]["index"]
        else:
            _msg = f"Unsupported model framework `{self.model_framework}` in breakover plugin."
            self.logger.error(_msg)
            raise Exception(_msg)

        self.logger.info(f"Model loaded from `{model_path}`")
        return model, input_details, output_details

    def predict(self, data_array):
        """Perform prediction.

        Args:
            data_array: a 3D input data array for performing prediction
                where (n_samples, sequence_length, n_features)
        Return:
            score: predicted score
            time_elapsed: inference time
        """

        t0 = time.time()

        if np.isnan(data_array).any():
            tmp = list(np.count_nonzero(np.count_nonzero(
                np.isnan(data_array), axis=0), axis=0))
            self.logger.warn(
                "In the input data array of shape `{data_array.shape}` the "
                "number of missing values by fields are `{tmp}`. This is "
                "unexpected, please check your south service to see if it's"
                "generating data adequately."
            )

        # Prediction
        if self.model_framework == "keras":
            scores = self.model.predict(data_array).ravel()

        elif self.model_framework == "tflite":

            scores = []
            for i in range(len(data_array)):

                # Invoke model on the input data
                self.model.set_tensor(
                    self.input_details, np.float32(data_array[i:i + 1]))
                self.model.invoke()

                # perform prediction
                single_prediction = self.model.get_tensor(self.output_details)
                single_prediction = float(np.squeeze(np.squeeze(
                    single_prediction, axis=0), axis=0))
                scores.append(single_prediction)

                del single_prediction

        score = float(scores[-1])
        if np.isnan(score):
            tmp = list(np.count_nonzero(np.count_nonzero(
                np.isnan(data_array), axis=0), axis=0))
            self.logger.warn(
                f"Score value is observed to be missing. Observed Scores: {scores}. "
                f"Observed missing rows in the input fields: {tmp}"
            )
            score = DEFAULT_SCORE
        time_elapsed = time.time() - t0

        del data_array

        return score, time_elapsed
