# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Plugin module to apply post processing to model score."""

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


import os
import sys
import time
import itertools
import numpy as np
from typing import Union
from collections import deque

from foglamp.plugins.filter.breakover.core.base import _LOGGER
from foglamp.plugins.filter.breakover.core.base import (
    DEFAULT_SCORE, DEFAULT_FORECAST,
    DEFAULT_FORECAST_STATE_NAME, DEFAULT_FORECAST_STATE_ID
)


class _Techniques:

    def __init__(
        self, score_deque, duration_deque,
        rolling_measures, rolling_window_length, rolling_measure_name,
        threshold, duration_cutoff
    ):
        """Accept params for post processing the score.

        The property of this class will return
            forecast (int) (-1/0/1)
            _msg     (str)

        Also, deque are passed by reference. So changes here will reflect to original.

        Args:
            score_deque: fixed size list containing raw scores (old to recent).
            duration_deque: fixed size list of size duration which contain the
                last inference forecasts (0/1).
            rolling_measures: PostProcess.measures (dict)
            rolling_window_length: window length on which the rolling measure
                will be calculated
            rolling_measure_name: the measure to be calculated on window.
            threshold: score threshold
            duration_cutoff: the duration for which the score is to remain above
                the score threshold
        """
        self.score_deque = score_deque
        self.duration_deque = duration_deque
        self.rolling_measures = rolling_measures
        self.rolling_window_length = rolling_window_length
        self.rolling_measure_name = rolling_measure_name
        self.threshold = threshold
        self.duration_cutoff = duration_cutoff

    @property
    def calculate_score_above_threshold(self):
        """Threshold the raw score."""
        recent_score = self.score_deque[-1]
        if recent_score >= self.threshold:
            forecast = 1
        else:
            forecast = 0

        _msg = f"recent value `{recent_score}`"
        del recent_score
        return forecast, _msg

    @property
    def calculate_rolling_measure_above_threshold(self):
        """Threshold the rolling measure calculated on raw score."""
        if (
            (len(self.score_deque) < self.rolling_window_length) or
            (DEFAULT_SCORE in self.score_deque)
        ):
            forecast = DEFAULT_FORECAST
            _msg = f"rolling value `{DEFAULT_SCORE}`"
        else:
            score_recent_to_old = list(
                itertools.islice(self.score_deque, 0, self.rolling_window_length)
            )[::-1]
            rolling_value = self.rolling_measures[self.rolling_measure_name](score_recent_to_old)
            if rolling_value >= self.threshold:
                forecast = 1
            else:
                forecast = 0

            _msg = f"rolling value `{rolling_value}`"
            del score_recent_to_old, rolling_value

        return forecast, _msg

    @property
    def calculate_score_above_threshold_for_a_duration(self):
        """Threshold based on duration with score above a threshold."""
        recent_score = self.score_deque[-1]
        true_duration = DEFAULT_SCORE

        if recent_score >= self.threshold:
            self.duration_deque.append(1)
        else:
            self.duration_deque.append(0)

        true_duration = sum(self.duration_deque)
        if len(self.duration_deque) < self.duration_cutoff:
            forecast = DEFAULT_FORECAST
        elif true_duration == self.duration_cutoff:
            forecast = 1
        else:
            forecast = 0

        _msg = f"recent value `{recent_score}` and continuous duration `{true_duration}`"
        del recent_score, true_duration

        return forecast, _msg

    @property
    def calculate_rolling_measure_above_threshold_for_a_duration(self):
        """Threshold the rolling measure on score and duration."""
        rolling_value = DEFAULT_SCORE
        true_duration = DEFAULT_SCORE

        if (
            (len(self.score_deque) < self.rolling_window_length) or
            (DEFAULT_SCORE in self.score_deque)
        ):
            forecast = DEFAULT_FORECAST
        else:
            score_recent_to_old = list(
                itertools.islice(self.score_deque, 0, self.rolling_window)
            )[::-1]
            rolling_value = self.rolling_measures[self.rolling_measure_name](score_recent_to_old)
            if rolling_value >= self.threshold:
                self.duration_deque.append(1)
            else:
                self.duration_deque.append(0)

            true_duration = sum(self.duration_deque)
            if len(self.duration_deque) < self.duration_cutoff:
                forecast = DEFAULT_FORECAST
            elif true_duration == self.duration_cutoff:
                forecast = 1
            else:
                forecast = 0

        _msg = f"rolling_value `{rolling_value}` and continuous duration `{true_duration}`"
        del score_recent_to_old, rolling_value, true_duration

        return forecast, _msg


class _TechniqueWrapper:

    @staticmethod
    def calculate_score_above_threshold(*args):
        return _Techniques(*args).calculate_score_above_threshold

    @staticmethod
    def calculate_rolling_measure_above_threshold(*args):
        return _Techniques(*args).calculate_rolling_measure_above_threshold

    @staticmethod
    def calculate_score_above_threshold_for_a_duration(*args):
        return _Techniques(*args).calculate_score_above_threshold_for_a_duration

    @staticmethod
    def calculate_rolling_measure_above_threshold_for_a_duration(*args):
        return _Techniques(*args).calculate_rolling_measure_above_threshold_for_a_duration


class PostProcess:
    """A class that apply post processing on the model score."""

    __slots__ = ["logger", "params", "score_deque", "duration_deque"]

    measures = {
        "Summation": np.sum,
        "Minimum": np.min,
        "Average": np.mean,
        "Median": np.median,
        "Maximum": np.max,

        "Standard Deviation": np.std,
        "Variance": np.var,
    }
    techniques = {
        "Score above threshold": _TechniqueWrapper.calculate_score_above_threshold,
        "Rolling measure above threshold": _TechniqueWrapper.calculate_rolling_measure_above_threshold,
        "Score above threshold for x duration": _TechniqueWrapper.calculate_score_above_threshold_for_a_duration,
        "Rolling measure above threshold for x duration": _TechniqueWrapper.calculate_rolling_measure_above_threshold_for_a_duration,
    }

    def __init__(
        self,
        params = [
            {
                "state_name": "positive",
                "technique": "Rolling measure above threshold",
                "rolling_window_length": 10,
                "rolling_measure_name": "Average",
                "threshold": 0.5,
                "duration_cutoff": None
            }
        ],
        logger = _LOGGER
    ):
        """Initialize Post Processing class.

        If -1 is observed in the data window then the prediction will be converted to -1.

        Args:
            params: a list containing the set info for the post processing
            logger: logger instance to use
        """
        self.logger = _LOGGER
        self.params = [
            {
                "state_name": d["state_name"],
                "state_id": d["state_id"],
                "technique": d["technique"],
                "rolling_window_length": d["rolling_window_length"],
                "rolling_measure_name": d["rolling_measure_name"],
                "threshold": d["threshold"],
                "duration_cutoff": d["duration_cutoff"],
            }
            for d in params
        ]

        # Initialing a common score queue that will keep track of the raw score
        # values from the model on which all the the post process operation
        # will be performed. This queue will be common for all set and will be
        # sliced according to set param.
        rolling_window_lengths = [
            d["rolling_window_length"]
            for d in params
            if d["rolling_window_length"] is not None
        ]
        if len(rolling_window_lengths) > 0:
            self.score_deque = deque(maxlen = max(rolling_window_lengths))
        else:
            self.score_deque = deque(maxlen = 1)

        duration_thresholds = [
            d["duration_cutoff"]
            for d in params
            if d["duration_cutoff"] is not None
        ]
        if len(duration_thresholds) > 0:
            self.duration_deque = deque(maxlen = max(duration_thresholds))
        else:
            self.duration_deque = None

    def register_score_value(self, curr_value: Union[int, float]):
        """Add value in the raw model score queue.

        Args:
            curr_value: Get the new reading for the series
        Return:

        """
        self.score_deque.append(curr_value)
        del curr_value

    def _generate_forecast(
        self, state_name, state_id, technique,
        rolling_window_length, rolling_measure_name, threshold, duration_cutoff
    ):
        """Calculate forecast on a single set of post processing operation.

        Args:
            set_name: set name for the post processing parameter combination.
            set_id: set id for the post processing parameter combination. This
                id will be mapped on HMI and will be used to indicate action.
            technique: the post processing technique to be applied
            rolling_window_length: window length on which the rolling measure
                will be calculated
            rolling_measure_name: the measure to be calculated on window.
            threshold: score threshold
            duration_cutoff: the duration for which the score is to remain above
                the score threshold
        Return:
            forecast: int 0/1; representing false/true
        """
        _msg = (
            "For `{}` ({}) with `{}` > rolling (measure `{}` & length `{}`)"
            " // threshold `{}` // duration_cutoff `{}`, "
        ).format(
            state_name, state_id, technique, rolling_measure_name,
            rolling_window_length, threshold, duration_cutoff
        )

        # Calculate measure and applying threshold
        forecast, _mssg = self.techniques[technique](
            self.score_deque, self.duration_deque, self.measures,
            rolling_window_length, rolling_measure_name,
            threshold, duration_cutoff
        )

        _msg += "of breakover is forecasted as `{}` i.e. {}. ({}).".format(
            forecast,
            (
                "breakover is probable"  # imminent, not used since warning zone is also available
                if forecast == 0 else
                (
                    "breakover is unlikely"
                    if forecast == 1 else
                    "breakover can't be predicted"
                )
            ),
            _mssg
        )
        _msg += " Size of deque (score/duration): `{} / {}`.".format(
            sys.getsizeof(self.score_deque),
            sys.getsizeof(self.duration_deque)
        )
        self.logger.debug(_msg)

        del _msg, _mssg

        return forecast

    def generate_forecast(self):
        """Generate forecast in accordance with post processing configuration.

        Return:
            a dict mapping state name and breakover forecast (-1/0/1)
                eg.
                    {
                        "healthy": 1,
                        "warning": 1,
                        "critical": 0
                    }
            a dict mapping the forecasted state
                eg.
                    {
                        "name": "warning",
                        "id": 1
                    }
        """
        state_forecast = [
            (d["state_name"], d["state_id"], self._generate_forecast(**d))
            for d in self.params
        ]

        # get all state which are forecasted as True/1
        states = [(t[0], t[1]) for t in state_forecast if t[-1] == 1]

        # Returning the predicted state; incase multiple states are forecasted,
        # select the one with the highest state_index
        forecasted_state = {
            "name": DEFAULT_FORECAST_STATE_NAME,
            "id": DEFAULT_FORECAST_STATE_ID
        }

        if len(states) > 0:
            state_name, state_id = max(states, key= lambda x: x[1])
            forecasted_state = {
                "name": state_name,
                "id": state_id
            }
            del state_name, state_id

        # Create state wise forecast
        # this will be returned to indicate forecast by state
        forecast_state_map = {
            state: forecast
            for (state, _, forecast) in state_forecast
        }

        return forecast_state_map, forecasted_state
