# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END


__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


import gc
import time
from typing import Union
from datetime import datetime, timezone
import numpy as np
import pandas as pd

from foglamp.plugins.filter.breakover.core.base import (
    _LOGGER, LOG_LEVEL, TriggerExecution, TrackResourceUsage,
    DEFAULT_SCORE, DEFAULT_FORECAST, DEFAULT_FORECAST_STATE_NAME,
    DEFAULT_FORECAST_STATE_ID
)
from foglamp.plugins.filter.breakover.config import parse_config
from foglamp.plugins.filter.breakover.core.bucket_handler import get_model_path
from foglamp.plugins.filter.breakover.core.data_handler import DataAccumulator
from foglamp.plugins.filter.breakover.core.inference_handler import Inference
from foglamp.plugins.filter.breakover.core.post_process_handler import PostProcess


class BreakoverExec:
    """Execute the  with FogLAMP."""

    __slots__ = [
        "add_metadata", "breakover_cfg", "data_manager", "enable", "logger",
        "predictor", "post_processor", "trigger_forecast", "trigger_gc",
        "last_poll_timestamp"
    ]

    def __init__(self, plugin_cfg, bucket_client):
        """Initialize breakover instance

        Args:
            plugin_cfg: filter configuration
            bucket_client: bucket client instance
        """
        self.breakover_cfg = parse_config(plugin_cfg)
        _LOGGER.info(f"breakover parsed config :  {self.breakover_cfg}")

        # re-adjust logging level & update config
        _LOGGER.setLevel(LOG_LEVEL[self.breakover_cfg["log_level"]])
        self.logger = _LOGGER

        self.breakover_cfg["inference"]["bucket"]["logger"] = self.logger
        self.breakover_cfg["inference"]["bucket"]["bucket_client"] = bucket_client

        self.add_metadata = self.breakover_cfg["add_metadata"]
        self.enable = self.breakover_cfg["enable"]
        self.logger.debug("Breakover Forecasting plugin is set to `Enable`.")

        if self.enable:
            # To enable unit test to send the model path, the following key model_path>value is added.
            self.initialize(model_path = plugin_cfg.get("model_path", {}).get("value"))

        self.last_poll_timestamp = None

    def _initialize_model(self, model_path = None):
        """ """
        # get model path; updated logger is also shared
        #     The model initialization wait and retry for the scenario where
        #     bucket service might not have initialized at time of this plugin
        #     initialization.
        sleep_duration = 2
        retry_counter, max_counter = 0, 30  # retry durations is 60 seconds
        while ((model_path is None) & (retry_counter < max_counter)):

            model_path = get_model_path(**self.breakover_cfg["inference"]["bucket"])

            if model_path is None:
                _msg  = "Unable to get model from bucket service. Retrying "
                _msg += f"[{retry_counter+1}/{max_counter}] Time passed "
                _msg += f"{sleep_duration*retry_counter} seconds."
                self.logger.info(_msg)
                time.sleep(sleep_duration)
                retry_counter += 1
        else:
            self.logger.info("Receive breakover model from bucket service.")

        del sleep_duration, retry_counter, max_counter

        if model_path is None:
            # unable to get model from bucket service & error has been raise
            self.enable = False
            _msg = "BreakoverExec.initialize() has failed as the plugin was "
            _msg += "unable to get model from bucket service. For now the "
            _msg += "the filter plugin will be diabled. Please check the bucket"
            _msg += "service and toggle breakover filter configuration `enable`"
            _msg += "to restart the plugin."
            self.logger.error(_msg)
            return

        # Initialize Inference; updated logger is also shared
        self.predictor = Inference(
            model_path = model_path,
            logger=self.logger
        )

    def initialize(self, model_path = None):
        """Initialize modules for enabling breakover forecasting."""
        # Initialize Data Preparation
        self.data_manager = DataAccumulator(
            self.breakover_cfg["inputs"],
            sequence_length=self.breakover_cfg["inference"]["sequence_length"],
            pad_sequence_length=5,
            logger=self.logger
        )

        # Initialize Inference Engine
        self._initialize_model(model_path = model_path)

        # Initialize Post Processor of Score
        self.post_processor = PostProcess(
            params = self.breakover_cfg["post_process"],
            logger = self.logger
        )

        # Initialize Trigger
        self.trigger_forecast = TriggerExecution(
            wait_period = self.breakover_cfg["inference"]["trigger_interval"]
        )

        self.logger.info("BreakoverExec.initialize() completed.")

    def get_logger(self):
        """Return the updated logger instance."""
        return self.logger

    def run(self, data: Union[list, dict]):
        """Execute breakover forecasting on new data.

        Args:
            data: readings data
        Return:
            data: updated data
        """
        if not self.enable:
            return

        iteration_start_ts = time.time()
        poll_interval = (
            (iteration_start_ts - self.last_poll_timestamp)
            if self.last_poll_timestamp is not None else
            -1.0
        )
        self.last_poll_timestamp = time.time()

        if isinstance(data, dict):
            data = [data]

        # add data for feature generation
        self.data_manager.filter_and_add_data(data)

        # calculate score if allowed & append to data
        core_reading = None
        score = DEFAULT_SCORE
        data_prep_time, inference_time, post_process_time = 0.0, 0.0, 0.0
        max_user_ts = self.data_manager.max_user_ts
        current_ts = pd.to_datetime(datetime.utcnow()).replace(tzinfo=timezone.utc)

        resource_tracker = TrackResourceUsage()
        with resource_tracker:

            if self.trigger_forecast.is_execution_allowed:

                self.trigger_forecast.reset_timer  # reset early, else inference time will get appended & delay will occur

                recent_row, latest_current_value = {}, None
                data_array = None
                forecasted_state = {
                    "name": DEFAULT_FORECAST_STATE_NAME,
                    "id": DEFAULT_FORECAST_STATE_ID
                }
                state_wise_forecast = {}

                try:
                    recent_row, data_array, max_user_ts, data_prep_time = self.data_manager.get_prepared_data()
                    latest_current_value = {
                        "current": val for k, val in recent_row.items()
                        if ((("current" in k.lower()) or ("amps" in k.lower())) and ~np.isnan(val))
                    }.get("current", None)

                except Exception as e:
                    _msg = f"Error in breakover data preparation. `{str(e)}`. Skipping inference generation..."
                    self.logger.error(_msg)

                post_process_start_ts = time.time()
                try:
                    if data_array is not None:
                        score, inference_time = self.predictor.predict(data_array)

                        ###########################################################
                        # HARDCODED LOGIC -------- MODIFY THIS LATER | TODO       #
                        # if current value is less than 2 Amps, set score to 0.0  #
                        if latest_current_value is not None:                      #
                            if latest_current_value < 2.0:                        #
                                score = 0.0                                       #
                        ###########################################################

                        # Adding reading for post processing & generating forecast
                        self.post_processor.register_score_value(score)
                        (
                            state_wise_forecast, forecasted_state
                        ) = self.post_processor.generate_forecast()

                except Exception as e:
                    _msg = "Error in breakover forecasting. `{}`".format(str(e))
                    self.logger.error(_msg)
                post_process_time = time.time() - post_process_start_ts
                del post_process_start_ts

                if data_array is not None:
                    state_name, forecast = None, None

                    core_reading = {
                        "readings": {
                            self.breakover_cfg["output"]["score_datapoint"]: score,
                        },
                        "asset": self.breakover_cfg["output"]["asset"],
                        "id": 0,
                        "ts": str(max_user_ts),
                        "user_ts": str(max_user_ts)       # what timestamp makes the most sense here; str(current_ts), str(max_user_ts)
                    }

                    if self.breakover_cfg["output"]["forecasted_state_datapoint"] is not None:
                        dp_name = self.breakover_cfg["output"]["forecasted_state_datapoint"]
                        key = self.breakover_cfg["output"]["return_state"]
                        core_reading["readings"][dp_name] = forecasted_state[key]
                        del dp_name, key

                    state_wise_forecast_prefix = self.breakover_cfg["output"]["state_wise_forecast_datapoint_prefix"]
                    if state_wise_forecast_prefix is not None:
                        for state_name, forecast in state_wise_forecast:
                            core_reading["readings"][f"{state_wise_forecast_prefix}{state_name}"] = forecast

                    del state_wise_forecast_prefix, state_name, forecast
                del (
                    data_array, forecasted_state, state_wise_forecast,
                    recent_row, latest_current_value
                )

            # forward core data
            if self.breakover_cfg["output"]["forward_data"]:
                if core_reading is not None:
                    data.append(core_reading)
                else:
                    pass
            else:
                if isinstance(core_reading, dict):
                    data = [core_reading]
                else:
                    data = []

        # prepare metadata & debug message
        resource_stats = resource_tracker.stats()
        iteration_duration = time.time() - iteration_start_ts
        _msg  = f"Prediction:: max_user_ts: {max_user_ts} "
        _msg += f"| score              : {score:.3f}\n"
        _msg += f"| poll interval      : {poll_interval:.3f} s "
        _msg += f"| iteration duration : {iteration_duration:.3f} s "
        self.logger.debug(_msg)

        metadata_reading = {}
        if self.add_metadata:
            metadata_reading = {
                "readings": {
                    "latency_at_filter_sec": (current_ts - max_user_ts).total_seconds(),
                    "poll_interval_sec": poll_interval,
                    "approx_iteration_time_sec": iteration_duration,
                    "duration_data_preparation_sec": data_prep_time,
                    "duration_inference_sec": inference_time,
                    "duration_post_process_sec": post_process_time,
                },
                "asset": "{}_metadata".format(self.breakover_cfg["output"]["asset"]),
                "id": 0,
                "ts": str(current_ts),
                "user_ts": str(current_ts)
            }
            for key, value in resource_stats.items():
                metadata_reading["readings"][key] = value
            data.append(metadata_reading)

        del (
            max_user_ts, score, core_reading, metadata_reading,
            resource_tracker, resource_stats,
            data_prep_time, inference_time, post_process_time,
            iteration_duration, iteration_start_ts, current_ts
        )

        _msg = f"A total of {len(data)} elements received in the data list."
        self.logger.debug(_msg)

        del _msg

        return data

    def stop(self):
        """Stop the breakover forecasting instance."""
        self.trigger_forecast.shutdown_timer

        self.trigger_forecast = None
        self.breakover_cfg = None
        self.data_manager = None
        self.predictor = None
        self.post_processor = None
